from copy import deepcopy
from itertools import combinations

from rummy.deck.sort import Sort


class Melds:

    def __init__(self):
        self.sort = Sort()
        self.melds = []

    def findDiscardScores(self, hand, discard=None):
        scores = []
        handCopy = deepcopy(hand)
        if discard:
            handCopy.append(discard)
        for i in range(8):
            dummyHand = deepcopy(handCopy)
            dummyHand.pop(i)
            score = self.calculateScore(dummyHand)
            scores.append(score)
        return scores

    def calculateScore(self, hand):
        self.melds = []
        hand = self.sort.sortHandBySuitAndRank(hand)
        cards = {self.sort.suitAndRankKey(card) for card in hand}
        self.findSets(hand)
        self.findRuns(hand)
        allPossibleMelds = self.findAllPossibleMelds()
        if len(allPossibleMelds) > 0:
            scores = self.findLowestScoringMelds(allPossibleMelds, cards)
            score = min(scores)
        else:
            score = sum([x[1] + 1 for x in cards])
        return score

    def findAllPossibleMelds(self):
        allPossibleMelds = []
        for L in range(1, 3):
            for subset in combinations(self.melds, L):
                allPossibleMelds.append(subset)
        return allPossibleMelds

    def findLowestScoringMelds(self, allPossibleMelds, cards):
        scores = []
        for item in allPossibleMelds:
            if len(item) > 1:
                for i in range(len(item) - 1):
                    if item[i].isdisjoint(item[i + 1]):
                        scores = self.findScores(item, i, cards, scores)
            else:
                remainingCards = cards.difference(item[0])
                scores.append(sum([x[1] + 1 for x in remainingCards]))
        return scores

    @staticmethod
    def findScores(item, i, cards, scores):
        if item[i].isdisjoint(item[i + 1]):
            items = item[i] | item[i + 1]
            remainingCards = cards.difference(items)
            scores.append(sum([x[1] + 1 for x in remainingCards]))
            return scores
        else:
            return []

    def findSets(self, hand):
        hand = self.sort.sortHandByRank(hand)
        cards = [self.sort.suitAndRankKey(card) for card in hand]
        i = 1
        while i < len(cards):
            i, meld = self.makeSetMeld(cards, i)
            self.makeAllMelds(meld)
            i += 1

    def findRuns(self, hand):
        hand = self.sort.sortHandBySuitAndRank(hand)
        cards = [self.sort.suitAndRankKey(card) for card in hand]
        i = 1
        while i < len(cards):
            i, meld = self.makeRunMeld(cards, i)
            self.makeAllMelds(meld)
            i += 1

    def makeSetMeld(self, cards, i):
        meld = []
        while i < len(cards) and cards[i][1] == cards[i - 1][1]:
            meld.append(cards[i - 1])
            i += 1
        meld.append(cards[i - 1])
        return i, meld

    def makeRunMeld(self, cards, i):
        meld = []
        while i < len(cards) and cards[i][0] == cards[i - 1][0] and cards[i][1] == cards[i - 1][1] + 1:
            meld.append(cards[i - 1])
            i += 1
        meld.append(cards[i - 1])
        return i, meld

    def makeAllMelds(self, meld):
        if len(meld) >= 3:
            self.melds.append(set(meld))
        if len(meld) > 3:
            for width in range(3, len(meld)):
                for i, step in enumerate(range(len(meld) - 2)):
                    self.melds.append(set(meld[step:width + i]))


if __name__ == '__main__':
    import sys


    def main():
        from rummy.deck.rank import Rank
        key = [
            'A♠', '2♠', '3♠', '4♠', '5♠', '6♠', '7♠',
            '8♠', '9♠', 'T♠', 'J♠', 'Q♠', 'K♠',
            'A♥', '2♥', '3♥', '4♥', '5♥', '6♥', '7♥',
            '8♥', '9♥', 'T♥', 'J♥', 'Q♥', 'K♥',
            'A♦', '2♦', '3♦', '4♦', '5♦', '6♦', '7♦',
            '8♦', '9♦', 'T♦', 'J♦', 'Q♦', 'K♦',
            'A♣', '2♣', '3♣', '4♣', '5♣', '6♣', '7♣',
            '8♣', '9♣', 'T♣', 'J♣', 'Q♣', 'K♣'
        ]
        rank = Rank()
        melds = Melds()
        cards = rank.rankedCards
        while True:
            handString = input('Hand String: ')
            hand = [cards[key.index(card)] for card in handString.split(', ')]
            if len(hand) > 7:
                scores = melds.findDiscardScores(hand)
                print('Possible Scores: ', scores)
                print('Discard: ', hand.pop(scores.index(min(scores))))
            print('Hand: ', ', '.join([str(card) for card in hand]))
            print('Score: ', melds.calculateScore(hand))
            print("\n==============================\n")


    try:
        main()
    except KeyboardInterrupt:
        sys.exit()



# -*- coding: utf-8 -*-
from rummy.deck.deck import Deck
from rummy.player.dealer import Dealer


class Round(Dealer, Deck):
    firstPlayer = 0
    currentPlayer = 0
    turn = 1
    lastTurn = 1
    knocked = False

    def __init__(self, players):
        super().__init__()
        self.players = players
        self.stackDeck()

    def prepareNewRound(self):
        self.turn = 1
        self.lastTurn = 1
        self.knocked = False
        self.stackDeck()

    def dealCards(self, players):
        for p in players:
            p.hand = self.deal(self.deck)

    def prepareTurn(self):
        self.checkStack()
        self.checkKnocked()

    def getTurn(self, round):
        return "Turn %i, %s\n" % (round.turn, round.getCurrentPlayersName())

    def getCurrentPlayersHand(self):
        return self.players[self.currentPlayer].getHandToPrint()

    def endTurn(self):
        self.switchCurrentPlayer()
        self.turn += 1

    def checkKnocked(self):
        if self.knocked:
            self.lastTurn += 1

    def switchCurrentPlayer(self):
        self.currentPlayer = (self.currentPlayer + 1) % len(self.players)

    def rotateFirstPlayer(self):
        self.firstPlayer += 1
        self.currentPlayer = self.firstPlayer % len(self.players)



# -*- coding: utf-8 -*-
from .card import Card
from ..unicode.check import hasUnicode


class Rank:
    values = ["A"] + [str(d) for d in list(range(2, 10))] + ["T", "J", "Q", "K"]
    suits = [u"\u2660", u"\u2665", u"\u2666", u"\u2663"]

    def __init__(self):
        if not hasUnicode:
            self.suits = ["S", "H", "D", "C"]
        self.rankedCards = [Card(value, suit) for suit in self.suits for value in self.values]

    def __str__(self):
        return str([(i, str(card)) for i, card in
                    enumerate([(Card(value, suit)) for suit in self.suits for value in self.values])])



# coding=utf-8
from random import choice
from time import sleep

from text_template import TextTemplate as View

from rummy.constants.package_resource_path import TEMPLATE_PATH
from rummy.player.player import Player


class AI(Player):

    def chooseToDiscardOrPickUp(self):
        if len(self.round.discard) > 0:
            if self.aiOnly:
                self.renderPlayerTurnStart()
                self.aiThinking('Choosing pick up')
            else:
                self.renderAITurnStart()
            self.choosePickUp()
        else:
            if self.aiOnly:
                self.renderPlayerTurnStart()
            else:
                self.renderAITurnStart()
            self.aiThinking('Drawing first card from deck')
            self.hand.drawCard(self.round.deck.pop())

    def choosePickUp(self):
        currentScore = self.hand.getScore()
        if self.aiOnly:
            print('Current Score: ', currentScore)
        scores = self.melds.findDiscardScores(self.hand.getHand(), self.round.discard[-1])
        if self.aiOnly:
            print('Possible Hand Scores: ', scores)
            print('Min Score: ', min(scores))
        if min(scores) < currentScore - 4 or min(scores) <= 10:
            self.aiThinking('Drawing from discard')
            self.hand.drawCard(self.round.discard.pop())
        else:
            self.aiThinking('Drawing from deck')
            self.hand.drawCard(self.round.deck.pop())

    def discardOrKnock(self):
        if self.aiOnly:
            self.renderAITurnEnd()
        self.aiThinking('Choosing card to discard')
        scores = self.melds.findDiscardScores(self.hand.getHand())
        score = min(scores)
        if score <= 10 and not self.round.knocked:
            self.round.knocked = True
        if scores.count(score) > 1:
            choices = [(i, x) for (i, x) in enumerate(scores) if (x == score)]
            discard = choice(choices)[0]
        else:
            discard = scores.index(score)
        discard = self.hand.discardCard(discard)
        if self.aiOnly:
            print('Discarding: ', discard)
            print('Hand Score: ', score)
        self.round.discard.append(discard)

    def aiThinking(self, action):
        print(View.render(
            template=TEMPLATE_PATH + '/ai-thinking.txt',
            action=action,
        ))
        if not self.aiOnly:
            sleep(0.8)

    def renderAITurnEnd(self):
        print(View.render(
            template=TEMPLATE_PATH + '/ai-turn-end.txt',
            hand=self.hand.getHandToPrint(),
        ))



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod

from text_template import TextTemplate as View

from rummy.constants.package_resource_path import TEMPLATE_PATH
from rummy.deck.melds import Melds
from rummy.player.hand import Hand


class Player(metaclass=ABCMeta):
    __metaclass__ = ABCMeta

    def __init__(self, num):
        self.num = num
        self.gameScore = 0
        self.hand = Hand()
        self.melds = Melds()

    def turn(self, round, aiOnly):
        self.round = round
        self.aiOnly = aiOnly
        self.hasSomeoneKnocked()
        self.chooseToDiscardOrPickUp()
        self.discardOrKnock()

    def getName(self):
        return "Player %i" % self.num

    def updateScore(self):
        self.gameScore += self.hand.getScore()

    def getHand(self):
        return self.hand

    def getGameScore(self):
        return self.gameScore

    def displayRoundScore(self):
        return self.hand.score

    def hasSomeoneKnocked(self):
        if self.round.knocked:
            print(View.render(template=TEMPLATE_PATH + '/knocked.txt'))

    def renderPlayerTurnStart(self):
        print(View.render(
            template=TEMPLATE_PATH + '/player-turn-start.txt',
            turn_number=self.round.turn,
            player_number=self.round.currentPlayer + 1,
            score=self.hand.getScore(),
            hand=self.hand.getHandToPrint(),
            discard=self.round.getDiscard()
        ))

    def renderAITurnStart(self):
        print(View.render(
            template=TEMPLATE_PATH + '/turn-start.txt',
            turn_number=self.round.turn,
            player_number=self.round.currentPlayer + 1,
            discard=self.round.getDiscard()
        ))

    def renderPlayerTurnEnd(self):
        print(View.render(
            template=TEMPLATE_PATH + '/player-turn-end.txt',
            hand=self.hand.getHandToPrint(),
            key=self.hand.getKey()
        ))

    @abstractmethod
    def chooseToDiscardOrPickUp(self):
        pass

    @abstractmethod
    def discardOrKnock(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Sep 25 14:14:18 2016

@author: Sarcoma
"""
import os
from time import sleep

import colorama

from rummy.game.round import Round
from rummy.game.score import Score
from rummy.game.setup_players import SetupPlayers
from rummy.player.human import Human
from ansi_colours import AnsiColours as Colour


class Play():
    def __init__(self):
        self.coloramaConfig()
        setup = SetupPlayers()
        self.players = setup.createPlayers()
        self.aiOnly = not any(isinstance(x, Human) for x in self.players)
        self.score = Score(self.players)
        self.round = Round(self.players)
        self.round.dealCards(self.players)
        self.playGame()

    def coloramaConfig(self):
        if 'PYCHARM_HOSTED' in os.environ:
            convert = False  # in PyCharm, we should disable convert
            strip = False
        else:
            convert = None
            strip = None
        colorama.init(convert=convert, strip=strip)

    def playGame(self):
        while self.round.lastTurn != len(self.players):
            self.round.prepareTurn()
            player = self.players[self.round.currentPlayer]
            player.turn(self.round, self.aiOnly)
            self.round.endTurn()
        self.endRound()
        sleep(1.2)
        self.startNewRoundOrEndGame()

    def startNewRoundOrEndGame(self):
        if self.score.isEndOfGame():
            self.score.endGame()
        else:
            self.round.rotateFirstPlayer()
            if not self.aiOnly:
                self.confirmStartNewRound()
            self.round.prepareNewRound()
            self.round.dealCards(self.players)
            self.playGame()

    def confirmStartNewRound(self):
        print("\nReady %s?" % self.players[self.round.currentPlayer].getName())
        ready = ''
        while ready.lower() != 'y':
            ready = input("Enter " + Colour.green('y') + " when you are ready for the next round: ")

    def endRound(self):
        self.score.updatePlayerScores()
        self.score.displayThisRoundScore()

# start game
if __name__ == "__main__":
    Play()



# -*- coding: utf-8 -*-
from rummy.player.hand import Hand


class Dealer:
    cardCount = 7

    def setCardCount(self, cardCount):
        self.cardCount = cardCount

    def deal(self, deck):
        hand = Hand()
        for _ in range(self.cardCount):
            hand.drawCard(deck.pop())
        return hand



# coding=utf-8
from time import sleep

from ansi_colours import AnsiColours as Colour
from text_template import TextTemplate as View

from rummy.constants.package_resource_path import TEMPLATE_PATH


class Score:
    def __init__(self, players):
        self.players = players

    def displayThisRoundScore(self):
        print(View.render(
            template=TEMPLATE_PATH + '/round-end.txt',
            round_scores=self.getEndOfRoundScores(),
            game_scores=self.getCurrentGameScores()
        ))

    def getCurrentGameScores(self):
        return ''.join(["%s: %s\n" % (p.getName(), p.getGameScore()) for p in self.players])

    def getEndOfRoundScores(self):
        output = ''
        lessThanTen = False
        for p in self.players:
            score = p.hand.getScore()
            if score <= 10:
                lessThanTen = True
            output += View.render(
                template=TEMPLATE_PATH + '/hand-score.txt',
                player=p.getName(),
                hand=p.hand.getHandToPrint(),
                score=score
            )
        if not lessThanTen:
            print(Colour.red('<+++++++++++++++++++BUG+++++++++++++++++++++++++'))
            sleep(8)
        return output

    def updatePlayerScores(self):
        for p in self.players:
            p.hand.getScore()
            p.updateScore()

    def isEndOfGame(self):
        for p in self.players:
            if p.getGameScore() >= 100:
                return True
        return False

    def endGame(self):
        winners = self.findLowestScores()
        if len(winners) == 1:
            print(Colour.green(winners[0].getName() + " is the Winner!!"))
        else:
            print(Colour.green(", ".join([w.getName() for w in winners]) + " are joint winners!"))

    def findLowestScores(self):
        lowest = []
        for p in self.players:
            if not lowest:
                lowest = [p]
                continue
            if p.getGameScore() < lowest[0].getGameScore():
                lowest = [p]
            elif p.getGameScore() == lowest[0].getGameScore():
                lowest.append(p)
        return lowest



# -*- coding: utf-8 -*-
from rummy.player.ai import AI
from rummy.player.human import Human


class SetupPlayers:
    players = []
    numberOfPlayers = -1
    numberOfOpponents = -1

    def __init__(self):
        self.choosePlayers()

    def choosePlayers(self):
        numberOfPlayers = -1
        while numberOfPlayers not in [i for i in range(0, 5)]:
            numberOfPlayers = input("Enter number of players (0-4)? ")
            numberOfPlayers = self.validNumberCheck(numberOfPlayers)
        if numberOfPlayers in [0, 1]:
            self.setupAI(numberOfPlayers)
        self.numberOfPlayers = numberOfPlayers

    def setupAI(self, numberOfPlayers):
        self.ai = True
        if numberOfPlayers == 0:
            self.chooseNumberOfAIOpponents(4)
        elif numberOfPlayers == 1:
            self.chooseNumberOfAIOpponents(3)

    def chooseNumberOfAIOpponents(self, maxOpponents):
        numberOfOpponents = -1
        while numberOfOpponents not in [i for i in range(maxOpponents - 2, maxOpponents + 1)]:
            numberOfOpponents = input("Enter number of opponents ({0}-{1})? ".format(maxOpponents - 2, maxOpponents))
            numberOfOpponents = self.validNumberCheck(numberOfOpponents)
        self.numberOfOpponents = numberOfOpponents

    @staticmethod
    def validNumberCheck(number):
        try:
            number = int(number)
        except ValueError:
            number = -1
            print("Not a valid number, please try again...")
        return number

    def createPlayers(self):
        i = 0
        players = []
        for j in range(self.numberOfPlayers):
            i += 1
            players.append(Human(i))
        for j in range(self.numberOfOpponents):
            i += 1
            players.append(AI(i))
        return players



# -*- coding: utf-8 -*-
from ansi_colours import AnsiColours as Colour

from ..unicode.check import hasUnicode


class Card:
    value = ""
    suit = ""

    def __init__(self, value, suit):
        self.value = value
        self.suit = suit

    def __str__(self):
        return "%s%s" % (self.value, self.suit)

    def getCardColour(self):
        if self.suit in [u"\u2665", u"\u2666", "H", "D"]:
            return self.redCard()
        elif self.suit in [u"\u2660", u"\u2663", "C", "S"]:
            return self.blackCard()

    def redCard(self):
        if hasUnicode:
            return str(self.value) + Colour.red(self.suit) + ", "
        else:
            return str(self.value) + self.suit + ", "

    def blackCard(self):
        return str(self.value) + self.suit + ", "



# -*- coding: utf-8 -*-

from ansi_colours import AnsiColours as Colour

from rummy.deck.melds import Melds
from rummy.deck.rank import Rank
from rummy.deck.sort import Sort


class Hand(Rank):

    def __init__(self):
        super().__init__()
        self.colour = Colour()
        self.hand = []
        self.score = 0
        self.sort = Sort()
        self.melds = Melds()

    def __str__(self):
        self.hand = self.sort.sortHandBySuitAndRank(self.hand)
        return ''.join([card.getCardColour() for card in self.hand]).strip(', ')

    def setHand(self, hand):
        self.hand = hand

    def getHand(self):
        return self.hand

    def drawCard(self, card):
        self.hand.append(card)

    def discardCard(self, choice):
        return self.hand.pop(choice)

    def getScore(self):
        return self.melds.calculateScore(self.hand)

    def getHandToPrint(self):
        self.hand = self.sort.sortHandBySuitAndRank(self.hand)
        return ''.join([card.getCardColour() for card in self.hand]).strip(', ')

    def getKey(self):
        return ''.join([" %s, " % Colour.green(str((i + 1))) for i in range(len(self.hand))])



# coding=utf-8
from ansi_colours import AnsiColours as Colour

from rummy.player.player import Player


class Human(Player):

    def chooseToDiscardOrPickUp(self):
        self.renderPlayerTurnStart()
        if len(self.round.discard) > 0:
            self.choosePickUp()
        else:
            self.hand.drawCard(self.round.deck.pop())

    def choosePickUp(self):
        playerChoice = self.getUserPickUpInput()
        if playerChoice == 'p':
            self.hand.drawCard(self.round.discard.pop())
        else:
            self.hand.drawCard(self.round.deck.pop())

    def discardOrKnock(self):
        self.renderPlayerTurnEnd()
        scores = self.melds.findDiscardScores(self.hand.getHand())
        if min(scores) <= 10 and not self.round.knocked:
            message = "Enter a number to discard a card or " + Colour.green('k') + " to Knock: "
        else:
            message = "Enter a number to discard a card: "
        playerChoice = ""
        while playerChoice not in [str(i) for i in range(1, 9)]:
            if self.round.knocked:
                message = "Enter a number to discard a card: "
            playerChoice = input(message)
            if playerChoice.lower() == "k" and min(scores) < 10:
                self.round.knocked = True
        playerChoice = int(playerChoice) - 1
        self.round.discard.append(self.hand.discardCard(playerChoice))

    def getUserPickUpInput(self):
        playerChoice = ''
        while playerChoice.lower() not in ['d', 'p']:
            playerChoice = input(
                "Enter " + Colour.green('d') + " to draw or " + Colour.green('p') + " to pickup discard: ")
        return playerChoice



# -*- coding: utf-8 -*-
from random import shuffle
from .rank import Rank

class Deck(Rank):
    deck = []
    discard = []

    def stackDeck(self):
        self.deck = self.rankedCards[:]
        self.discard = []
        shuffle(self.deck)

    def checkStack(self):
        if len(self.deck) == 0:
            self.deck = self.discard
            self.discard = []

    def getDiscard(self):
        return self.discard[-1].getCardColour() if len(self.discard) else 'Empty'



from rummy.deck.rank import Rank


class Sort:

    def __init__(self):
        self.rank = Rank()

    def sortHandBySuitAndRank(self, cards):
        return sorted(cards, key=lambda card: self.suitAndRankKey(card))

    def sortHandByRank(self, cards):
        return sorted(cards, key=lambda card: self.rankKey(card))

    def suitAndRankKey(self, card):
        return self.rank.suits.index(card.suit), self.rank.values.index(card.value)

    def rankKey(self, card):
        return self.rank.values.index(card.value)
